--
--   Version 1.0
--
--   Copyright 2025 Lee Djavaherian
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program.  If not, see <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------------------

--#########################################
--######## USER-PROVIDED CONSTANTS ########
--#########################################

HOMEPATH = "/index.gmi"

MIME = {gmi="text/gemini; charset=utf-8",
        txt="text/plain; charset=utf-8",
        jpg="image/jpeg",
        gif="image/gif",
        png="image/png",
        pdf="application/pdf"}

NOMIME = "text/gemini; charset=utf-8"

MAXTHREADS = 100
MAXPATHLEN = 400 --usually up to 4095 (path) + 255 (filename) - 1 = 4349 but you can cut this down further
MAXREQBYTES = 500 -- should be bigger than MAXPATHLEN
DOWNLOADBUFFER = 16384  -- 16K of RAM chunk per coroutine for big downloads, bypassed for .gmi files
MAXUPLOADBYTES = 100
TIMEOUTCHECKSECS = 5
TIMEOUTSECS = 60
DEBUG = false
UPLOADSENABLED = false
SAFEFILES = true -- requires Linux "find" command
RELOADPATH = "/reloadme" -- change this to something unique if SAFEFILES is true
STRICTMODE = false 
-- strict mode disables all chars except for alphanumeric, underscore, hyphen
-- forces 3 character extension, disables percent-decoding,
-- only allows alphanumeric and space in uploads, only allows one subdirectory level,
-- this tightens the filter and has lots of yielding/timeouts rather than rejections

--########################################
--### CONSTANT AND VARIABLE PROCESSING ###
--########################################

if not arg[1] or not arg[2] then
   print ("This is the Louia Spartan protocol TCP server.\nUsage: luajit louia.lua domain[:port] /sitedir [/logdir]")
   return
end
_, _, domain, port = string.find(arg[1],"([%w%-%.]+):?(%d*)")

domain = domain:gsub("([%.%-%[%]])", "%%%1") --escape periods/hyphens for ipv4, brackets for ipv6
if not port or port == "" then
  port = 300
end

dir = arg[2] -- example /mysite
logdir = arg[3] -- leave blank to disable logging, example /mylogdir

-- this protects again int overflow later by filtering out giant integers when in string form
if MAXUPLOADBYTES > 0 then
  maxuploaddigits = math.floor(math.log10(MAXUPLOADBYTES) + 1)
else 
  maxuploaddigits = 1
end

-- Five captures are: host, path, contentlength, upload
if STRICTMODE then 
   pattern = '^('..domain..')%s/([%w%%-%_]*/?[%w%%-%_]*%.?%w?%w?%w?)%s(%d+)\r\n([%w%s]*)$'
else
   pattern = '^('..domain..')%s/([%/%w%.%-%_%~%%%$%+%*%(%)%!\'%,%;]*)%s(%d+)\r\n(.*)$'
end
if DEBUG then print ("pattern is: " .. pattern) end

if SAFEFILES then
   -- load an array with all of the files in dir and subdirectories
   local myfiles = io.popen('find "'..dir..'" -type f')
   filearray = {}
   for line in myfiles:lines() do
     table.insert (filearray, line);
   end
   myfiles:close()
end

--##########################
--### COROUTINE FUNCTION ###
--##########################

function getpage (client)
  if DEBUG then print ("New coroutine started.") end
  local request = ""

  while true do

    local _, err, partial = client:receive(MAXREQBYTES) --don't need the first variable as that is for successful which cannot occur since I limit by length

    -- err has 3 responses: nil if MAXREQBYTES reached, closed if client closed early, timeout if not done
    
    if err == "timeout" then
       if partial then 
         request = request .. partial -- build up the full request string
       end

      local _, endmatch, host, path, contentlength, upload = string.find(request, pattern)

      if endmatch then

         if path == "" or path == nil then
            path = HOMEPATH
         else
            path = "/" .. path
         end
   
         local thedate = os.date()
         print (thedate .. " request received for " .. client:getpeername())
       
         -- ####### LOGGING ######
         if logdir then
           local flog = assert(io.open(logdir .. "/spartan.log", "a", "Error - Could not open logfile for appending."))
           -- uploads are only logged if in STRICTMODE and are alphanumeric with spaces only
           local tlog
           if STRICTMODE then
              tlog = flog:write(thedate .. " request received for " .. client:getpeername() .. ": " .. host .. " " .. path .. " " .. contentlength .. "\n" .. upload .. "\n")
           else
              tlog = flog:write(thedate .. " request received for " .. client:getpeername() .. ": " .. host .. " " .. path .. " " .. contentlength .. "\nUPLOADS NOT LOGGED")
           end
           flog:close()
         end

         local _, _, extension = string.find(path,"%.(%w*)$")  

         -- unfiltered request and upload strings are not printed for safety
         if DEBUG then 
           print ("request length: " .. endmatch)  
           print("host: " .. host) 
           print ("path without percent-decoding: " .. path)
           if extension then print ("extension: " .. extension) end
           print("content-length: " .. contentlength)
           print("actual upload size: " .. string.len(upload))
           print("maxuploaddigits: " .. maxuploaddigits)
         end

         if not STRICTMODE then
            -- percent-decode the path, except for bad chars
            -- there is no good way to break out of the gsub early so it has to run through
            local badchars = false
            path = string.gsub(path, "%%(%x%x)", 
                   function (n)
                      local num = tonumber(n, 16)
                      --sticks to ASCII + extra 8 bit chars, no control chars, no periods (46),
                      --and no slashes (47) no backslashes (92), no delete (127)
                      --this also blocks char 27 (ESC) so ANSI CSI commands cannot pass
                      if num < 31 or num == 46 or num == 47 or num == 92 or num == 127 then  
                         badchars = true
                         return "X"
                      else
                         return string.char(num) 
                      end
                   end)

            if badchars then
               client:send("4 Malformed request.  Bye.\r\n")
               print("Dangerous percent-encoded chars in path.")
               break
            end   

            if DEBUG then print ("path after percent-decoding: " .. path) end
         end 

         -- Now error out if bad stuff made it this far
         -- Lua uses short-cut or evaluation so extra checks are not costly
         -- The check for .. is redundant for STRICTMODE, but good here just in case
         if string.find(path, "//") or string.find(path, "%.%.") or
            string.find(path, "/$") or string.find(path, "/%.") or
            string.find(path, "%./") or extension == "lua" then
              client:send("4 Malformed request.  Bye.\r\n")
              print("Multiple slashes or possible attempt to traverse or access hidden file or .lua code.")
              break
         end
         
         -- ######### THIS IS A NORMAL DOWNLOAD REQUEST ##########
         if contentlength == "0" then

            if DEBUG then print ("opening dir is: " .. dir) end

            if string.len(path) > MAXPATHLEN then
              client:send("5 Path too long.  Bye.\r\n")
              print("Path length greater than MAXPATHLEN.")
              break
            end



            local f = false

            if SAFEFILES then

               if RELOADPATH == path then 
                  -- reload file array with all of the files in dir and subdirectories
                  local myfiles = io.popen('find "'..dir..'" -type f')
                  filearray = {}
                  for line in myfiles:lines() do
                    table.insert (filearray, line);
                  end
                  myfiles:close()
                  if DEBUG then print("File array reload requested and loaded.") end
                  client:send("2 text/gemini\r\nFile array has been reloaded.")
                  break
                end

               -- check safe array for file
               for _, file in ipairs(filearray) do
                 if file == dir .. path then
                   f=io.open(dir .. path, "r")
                   break 
                 end     
               end
            else
               -- test open read only, do not use assert here since we do not want to present error
               f=io.open(dir .. path, "r")
            end

  
            -- if file exists then proceed to check extensions and MIME types
            if f then
              if MIME[extension] then --valid extension exists
                 if extension == "gmi" then
                    if DEBUG then print("success, reading entire .gmi file at once") end
                    client:send("2 " .. MIME[extension] .. "\r\n" .. f:read("*a"))
                    f:close()
                    break
                 else
                    if DEBUG then print("success, non-gmi file") end
                    client:send("2 " .. MIME[extension] .. "\r\n")
                 end
              else
                if STRICTMODE then
                   client:send("5 File not found.  Bye.\r\n")
                   if DEBUG then print("File type not found.  Not serving it since in strictmode.") end
                   --don't do an f:close here or it will crash since not needed
                   break
                else
                   if DEBUG then print("success, but no extension or no MIME was found") end
                   client:send("2 " .. NOMIME .. "\r\n") 
                end
              end

              if DEBUG then print("reading in chunks") end
              -- read file and send in chucks to not overload RAM for large files
              while true do
                local chunk = f:read(DOWNLOADBUFFER)
                if not chunk then
                   break 
                end                
                client:send(chunk)  
              end
              f:close()

            else
              client:send("5 File not found.  Bye.\r\n")
              if DEBUG then print("File not found.") end
              --don't do an f:close here or it will crash since not needed
            end
            break
         
         -- ######## THIS IS AN UPLOAD REQUEST ########  
         elseif UPLOADSENABLED then           
            -- don't even attempt to assign int value yet but look at number of string digits first
            if string.len(contentlength) <= maxuploaddigits then
               local contentlengthnum = tonumber(contentlength)
               if contentlengthnum < MAXUPLOADBYTES then
                 local uploadlen = string.len(upload)
                 if uploadlen < contentlengthnum then
                    --string found but does not match yet        
                    if DEBUG then print ("Yielding.  Found upload but not enough bytes yet.") end
                    coroutine.yield(client)
                 elseif uploadlen == contentlengthnum then

                    --only small uploads work and is stored in variable "upload", you can add code here later
                    --if contentlengthnum > DOWNLOADBUFFER, is good idea to write smaller request chunks
                    --to a file rather than fill up memory, but this will have to be done above

                    client:send("2 text/gemini\r\nUpload received.")
                    print ("Upload received but discarded.")
                    -- Don't print upload string unless is only alphanumeric (STRICTMODE is on)
                    if DEBUG and STRICTMODE then print ("Upload is: " .. upload) end
                    break
                 else 
                    client:send("5 Upload too large.  Bye.\r\n")
                    print ("Upload too large, exceeded content-length parameter.")
                    break
                 end
               else
                  client:send("5 Upload too large.  Bye.\r\n")
                  print ("Upload too large, exceeded MAXUPLOADBYTES")
                  break
               end
            else 
               client:send("5 Upload too large.  Bye.\r\n")
               print ("Upload too large, exceeded MAXUPLOADDIGITS.")
               break
            end

         else
            client:send("5 Uploads not allowed at this time.  Bye.\r\n")
            print ("Uploads disabled.")
            break
         end   
 
      -- ####### TIMED OUT AND DOES NOT MATCH ANY REQUEST YET ########
      else     
        -- note that it will NEVER match above unless domain matches exactly along with other syntax       
         if DEBUG then print ("Yielding since no match found yet.") end
         coroutine.yield(client)
      end
    
    -- ########## CLIENT CLOSED EARLY #########   
    elseif err == "closed" then
      print ("Client closed early.")
      break

    -- ########## REQUEST REACHED MAX SIZE AND CLOSED NORMALLY, SO IS TOO LARGE #########     
    elseif err == nil then
         client:send("5 Request or upload string too large.  Bye.\r\n")
         print("Request or upload string too large, reached MAXREQBYTES.")
         break

    -- ######### CATCH ALL FOR UNKNOWN ERROR BUT SHOULD NEVER OCCUR #########
    else
       client:send("5 Unknown error.  Bye.\r\n")
       print ("Unknown error: " .. err)
       break
    end

  end -- end of while loop
  client:close()
end -- end function

-- ################################
-- ########## MAIN LOOP ###########
-- ################################
local socket = require("socket")
local server = assert(socket.bind("*", port), "Error - could not bind network port to server.") 
server:settimeout(0) --remove blocking on server
local threads = {}
local scanlist = {}
local n = 0

print("Louia server started on port " .. port .. " for domain: " .. domain)
print('To test, use: "telnet ' .. domain .. ' ' .. port .. '" then type "' .. domain ..  '/ 0"')
while true do

    table.insert(scanlist, server) -- add server object to existing list to catch new connections
    -- waiting for new connection or change in current connections
    socket.select(scanlist, nil, TIMEOUTCHECKSECS) --default blocking, but timeout to allow other timeouts
    scanlist = {} --clear table after there is a change and loop resumes

    local sclient = server:accept() --since non blocking, this may not be real
    -- if this is a real new connection and not nil, then create new coroutine

    if sclient then 
      if n > MAXTHREADS then
          sclient:send("5 Server exceeded.  Please try later.\r\n")
          print ("Server capacity exceeded.")
          sclient:close()
      else
          sclient:settimeout(0) -- make the new connection non blocking, too
          local co = coroutine.create(function () getpage(sclient) end)
          table.insert(threads, co)
          if DEBUG then print ("new coroutine created") end
      end
    end 

    n = table.getn(threads)
    if DEBUG then print ("number of coroutines is: " .. n) end
    --does not run if there are no threads and n is 0
    local i = 1
    while i <= n do 
      
      if DEBUG then print ("inside dispatch loop: " .. i) end
      local rstatus, rclient = coroutine.resume(threads[i]) --rclient in second value is returned by yield
                                                         
      if not rstatus then
         print ("Internal Error.  Coroutine " .. i .. " crashed with: " .. rclient)
         i = i - 1 --table shrunk to left
         n = n - 1 --table shrunk to left
         break
      -- if no couroutine crash, proceed on
      elseif not rclient then    -- if yield did not send rclient back, coroutine ended
        table.remove(threads, i)
        if DEBUG then print("removing thread") end
        i = i - 1 --table shrunk to left
        n = n - 1 --table shrunk to left
        print ("number of coroutines is: " .. n)
      else    -- rclient timeout and still in progress.  Then check my own timeout, too.
        local _,_,elapsed = rclient:getstats()
        if elapsed > TIMEOUTSECS then
          rclient:send("4 Client timeout.  Bye.\r\n")
          print ("Client timeout.")
          rclient:close()
          table.remove(threads, i)
          i = i - 1 --table shrunk to left
          n = n - 1 --table shrunk to left
        else
          if DEBUG then print("inserting yielded client object to scan list") end
          table.insert(scanlist, rclient)
        end
      end 
      if DEBUG then print ("size of scan list: " .. table.getn(scanlist)) end
    i = i + 1 -- increment thread position
    end
end
 
